In order to make a new type of data available in the memory system, you first need to create an ARON XML file. Please refer to [Aron/CodeGeneration](Aron/CodeGeneration) if you want to know how to do so.

[[_TOC_]]

<!--
# Define your data type in ARON

Consider you want to use a data type called `MyData` (in the namespace `armarx::mydata`) in the memory system. First, you need to describe it in ARON.

- Choose a library in an ArmarX package where your data type will be defined (or create a new one). Here, we will call it `MyDataLib`.
- In the directory `MyDataLib/`, create a directory `aron/` (if necessary). Inside, add a file `MyData.xml`. The directory structure should now look like this:

```plaintext
.../libraries/
- MyDataLib/
  - CMakeLists.txt
  - ...
  - aron/
    - MyData.xml
```

- In the `CMakeLists.txt`, add or extend after the definition of the target (e.g. through add_library or add_component):

```cmake
armarx_enable_aron_file_generation_for_target(
    TARGET_NAME
        ${LIB_NAME}
    ARON_FILES
        aron/MyData.xml
)
```

- Define your data in ARON. A hello world example could look like this:

```xml
<?xml version="1.0" encoding="UTF-8" ?>
<AronTypeDefinition>
    <CodeIncludes>
    </CodeIncludes>
    <AronIncludes>
    </AronIncludes>
    <GenerateTypes>

        <Object name="armarx::mydata::arondto::MyData">

            <ObjectChild key="helloWorld">
                <String />
            </ObjectChild>

        </Object>

    </GenerateTypes>
</AronTypeDefinition>
```

Note:

- `<Object>` defines a new ARON object type (generating a C++ class) containing a number of children.
- `<ObjectChild>` defines a child (generating a public member variable of the generated C++ class, with `key` specifying the member's name).
- `<String />` inside a `<ObjectChild>` defines the type of the child. It can be a number of primitive types, or fully qualified names of other ARON objects.

Conventions:

- Put your ARON object type into a `arondto::` namespace inside your usual namespace (e.g., `armarx::mydata::arondto::MyData`), where "DTO" stands for "Data Transfer Object". This way, you can define or use a custom or existing C++ type with more methods/intelligence in the usual namespace in your business logic code (e.g. `armarx::mydata::MyData`).

Here is an example of an ARON object representing the business object (BO) type `simox::OrientedBox`:

```xml
<?xml version="1.0" encoding="UTF-8" ?>
<AronTypeDefinition>
    <CodeIncludes>
        <Include include="<Eigen/Core>" />
        <Include include="<Eigen/Geometry>" />
    </CodeIncludes>
    <GenerateTypes>

         <Object name="simox::arondto::OrientedBox">
            <ObjectChild key='center'>
                <Position />
            </ObjectChild>
            <ObjectChild key='orientation'>
                <Orientation />
            </ObjectChild>
            <ObjectChild key='extents'>
                <Position />
            </ObjectChild>

        </Object>

    </GenerateTypes>
</AronTypeDefinition>
```

Note:

- (At the moment), you need to specify `<CodeIncludes>` to Eigen to use `Pose`, `Position` and `Orientation` (which are translated to `Eigen::Matrix4f`, `Eigen::Vector3f`, and `Eigen::Quaternionf` in C++).

If you want to use an ARON object type defined in another file, this could look like this:

```xml
<?xml version="1.0" encoding="UTF-8" ?>
<AronTypeDefinition>
    <CodeIncludes>
        <Include include="<RobotAPI/libraries/aron/common/aron/AxisAlignedBoundingBox.aron.generated.h>" />
        <Include include="<RobotAPI/libraries/aron/common/aron/OrientedBox.aron.generated.h>" />
        <Include include="<RobotAPI/libraries/aron/common/aron/PackagePath.aron.generated.h>" />
        <Include include="<RobotAPI/libraries/ArmarXObjects/aron/ObjectID.aron.generated.h>" />
        <Include include="<RobotAPI/libraries/ArmarXObjects/aron/ObjectNames.aron.generated.h>" />
        <Include include="<RobotAPI/libraries/armem/aron/MemoryID.aron.generated.h>" />
    </CodeIncludes>
    <AronIncludes>
        <Include include="<RobotAPI/libraries/aron/common/aron/AxisAlignedBoundingBox.xml>" />
        <Include include="<RobotAPI/libraries/aron/common/aron/OrientedBox.xml>" />
        <Include include="<RobotAPI/libraries/aron/common/aron/PackagePath.xml>" />
        <Include include="<RobotAPI/libraries/ArmarXObjects/aron/ObjectID.xml>" />
        <Include include="<RobotAPI/libraries/ArmarXObjects/aron/ObjectNames.xml>" />
        <Include include="<RobotAPI/libraries/armem/aron/MemoryID.xml>" />
    </AronIncludes>
    <GenerateTypes>

        <Object name="armarx::armem::arondto::ObjectClass">

            <ObjectChild key="id">
                <armarx::arondto::ObjectID />
            </ObjectChild>

            <ObjectChild key="simoxXmlPath">
                <armarx::arondto::PackagePath />
            </ObjectChild>

            <ObjectChild key="meshWrlPath">
                <armarx::arondto::PackagePath />
            </ObjectChild>

            <ObjectChild key="meshObjPath">
                <armarx::arondto::PackagePath />
            </ObjectChild>

            <ObjectChild key="aabb">
                <simox::arondto::AxisAlignedBoundingBox />
            </ObjectChild>
            <ObjectChild key="oobb">
                <simox::arondto::OrientedBox />
            </ObjectChild>

            <ObjectChild key="names">
                <armarx::arondto::ObjectNames />
            </ObjectChild>

        </Object>

    </GenerateTypes>
</AronTypeDefinition>
```

Note:

- You need to include other ARON files in `<AronIncludes>`. The paths have the same form as usual C++ includes and point to the ARON XML files.
- At the moment, you also have to `<CodeInclude>` the respective generated headers (`.../MyData.aron.generated.h`). These should be generated automatically in the future.

# (Optional) Add conversions to existing (potentially more intelligent) C++ types

Consider the case of `simox::OrientedBoxf` and `simox::arondto::OrientedBox`. The former is the business object (BO), used in daily business logic code (as it offers useful methods to construct and manipulate it). The latter is a mere data storage used to transfer the data, which is called a data transfer object (DTO).

Therefore, you usually want to convert a `simox::arondto::OrientedBox` to a `simox::OrientedBoxf` as soon as you want to do things other than passing it around over network (e.g., at the beginning of a component method or after reading it from the working memory). Likewise, you want to convert a `simox::OrientedBoxf` to a `simox::arondto::OrientedBox` when you send the information over network or store it in the memory.

To allow this conversion, follow these steps:

- In your library, add a file pair `aron_conversions.{h, cpp}` next to your `aron/` directory and add them to the `CMakeLists.txt`
- In the header, declare functions following this form:

```cpp
#pragma once

#include <SimoxUtility/shapes/OrientedBox.h>
#include <RobotAPI/libraries/aron/common/aron/OrientedBox.aron.generated.h>


namespace simox
{
    void fromAron(const arondto::OrientedBox& dto, OrientedBoxf& bo);
    void toAron(arondto::OrientedBox& dto, const OrientedBoxf& bo);
}
```

Note:

- Both functions take the DTO first and the BO second.
- `fromAron()` takes a `const` DTO and a `non-const` BO.
- `toAron()` takes a `non-const` DTO and a `const` BO.
- All arguments are passed as reference (especially the `non-const` arguments).
- The functions are defined in the BO's namespace (`simox` in this example, `armarx::mydata` in the running example).


* In the cpp file, define the functions like this:

```cpp
#include "aron_conversions.h"

// Include this and use armarx::aron::to/fromAron() for std::vector, std::map.
#include <RobotAPI/libraries/aron/common/aron_conversions.h>

// Include other aron_conversions.h files for other ARON types


void simox::fromAron(const arondto::OrientedBox& dto, OrientedBoxf& bo)
{
    bo = OrientedBoxf(dto.center, dto.orientation, dto.extents);
}

void simox::toAron(arondto::OrientedBox& dto, const OrientedBoxf& bo)
{
    dto.center = bo.center();
    dto.orientation = bo.rotation();
    dto.extents = bo.dimensions();
}
```

Note:

- The implementation depends on your data. It might be as simple as copying all members, or require more complex conversions. Especially, you might use `fromAron()` for `toAron()` for other ARON object types.

-->

# Create a new memory server

> **Note:** If you plan to extend an existing memory server, you can skip this step.

If your data should be stored in a new memory server, follow these steps:

Create a new component, let's say `MyMemory`:

```plaintext
# In the root directory of your ArmarX package:
armarx-package add component MyMemory
```

Then, extend the new component by the respective component plugin.

Add the include:

```cpp
#include <RobotAPI/libraries/armem/server/plugins/ReadWritePluginUser.h>

// Or:
#include <RobotAPI/libraries/armem/server/plugins/ReadOnlyPluginUser.h>
```

Make the component derive from the component plugin user:

```cpp
    class MyMemory :
        virtual public armarx::Component
        // v Add this line ... v
        , virtual public armarx::armem::server::ReadWritePluginUser
        // v ... or this line ... v
        , virtual public armem::server::ReadOnlyPluginUser
```

Don't forget to add the respective library to your `CMakeLists.txt`:

```cmake
set(COMPONENT_LIBS
    ...
    # Add this:
    armem 
    ...
)
```

**If the component already implements an ice interface,** you have to add the memory server into face to it (otherwise, you will receive `no unique final overrider for 'virtual function ...'` errors by the compiler):
```cpp
// memory_tutorial/source/memory_tutorial/components/object_memory/ComponentInterface.ice

#pragma once

// v  either include this one (read-write)  v
#include <RobotAPI/interface/armem/server/MemoryInterface.ice>
// v  or this one (read-only)  v
// #include <RobotAPI/interface/armem/server/ReadingMemoryInterface.ice>


module memory_tutorial {  module components {  module object_memory 
{

    interface ComponentInterface 
        // v  either extend from this one (read-write)  v
        extends armarx::armem::server::MemoryInterface
        // v  or this one (read-only)  v
        // extends armarx::armem::server::ReadingMemoryInterface
    {
        ...
    };

};};};
```

In this case, make sure `RobotAPIInterfaces` is in the Ice dependencies of your ice library:
```cmake
armarx_add_component(object_memory
    ICE_FILES
        ComponentInterface.ice
    ICE_DEPENDENCIES
        ...
        RobotAPIInterfaces
    ...
```


You can set the memory name of the memory server in the `createPropertyDefinitions()`:

```cpp
#include <RobotAPI/libraries/armem/server/wm/memory_definitions.h>

...

    armarx::PropertyDefinitionsPtr MyMemory::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr defs = new ComponentPropertyDefinitions(getConfigIdentifier());
        ...

        // Set the memory name:
        workingMemory().name() = "My";

        ...
        return defs;
    }
```

**Note:**

* The `ReadWritePluginUser` implements the ice interface `MemoryInterface`, which combines, among others, the `ReadingMemoryInterface` and the `WritingMemoryInterface`.
* The `ReadOnlyPluginUser` only implements the `ReadingMemoryInterface`.
* Both provide a working memory data storage `workingMemory()` (of type `armem::server::wm::Memory`) and an `iceAdapter()` (`armem::server::MemoryToIceAdapter`).

That's it: Your component is now a memory server.

# Add a core segment with your new data type in a memory server

> Wait, how does the new memory server know of my specially written ARON type?

Oh right, it doesn't. At least not yet. But this is easy to fix: We just have to add a core segment.

> What on earth is a core segment?

A _segment_ is a section of a memory (server) containing data of a specific type, i.e. they are homogeneous data containers (in contrast to heterogeneous, which would mean they contain different kinds of data) (see the [Introduction](ArMem/Introduction)).

There are two kinds of segments: _core_ segments, and _provider_ segments:

* **Core segments** are usually added by the memory server itself with a name and the ARON data type the core segment is going to keep.
* **Provider segments** are usually added by clients committing data to the memory. You can think of them as "sub-segments" of a (core) segment, providing a namespace for the provider's entities and some additional features.

> ... what?

Just stick with me, you'll get used to it.

> Okay ... but how does the memory add the ... core segment?

A fine question! Luckily, that's very simple. All we need is your ARON type, a name and the working memory data structure.

> ...

Okay, okay, I'll come to the code. Adding a core segment merely alters a local data structure (it is not an ice/network operation). Thus, we can do it in `onInitComponent()`:

```cpp
    void MyMemory::onInitComponent()
    {
        workingMemory().addCoreSegment("Data", armarx::mydata::arondto::MyData::toAronType());
    }
```

* `workingMemory()` is a function provided by the component plugin user (`ReadWritePluginUser` or `ReadOnlyPluginUser`) returning a reference to the working memory data of the type `armem::server::wm::Memory`.
* `addCoreSegment()` adds a core segment. It takes a name and (optionally) an ARON type.
* `"Data"` is the name of the core segment.
  * It should refer to a modality or concept, not to its elements (i.e. prefer "Instance" over "Instances" and "Location" over "Locations").
  * Also, try to avoid repeating the memory name, as the core segment's ID is prepended by the memory name anyway (e.g. "Object/Instance" over "Object/ObjectInstance" and "Navigation/Location" over "Navigation/NavigationLocation").
* `armarx::mydata::arondto::MyData` is the C++ class generated from your `MyData.xml` ARON XML description file.
  * `::toAronType()` creates a runtime type object describing the ARON type. This allows, e.g., the `MemoryViewer` GUI to show the data in a suitable manner (e.g. a pose as an affine transformation instead of a 4x4 matrix).
  * Of course, you need to include the respective header: It's located at the same path where your XML is. For example, for the `ObjectInstance.xml` located at `RobotAPI/libraries/armem_objects/aron/ObjectInstance.xml`, the include would be:

```cpp
#include <RobotAPI/libraries/armem_objects/aron/ObjectInstance.h>
```

Note that you need to include the `.h` file, not the `.xml`. As the generated class is header-only, you do not need to link any libraries in your `CMakeLists.txt`.

When you now start your new memory server and the MemoryNameSystem (MNS) (e.g. using the scenario _ArMemCore_ in RobotAPI), and open the _MemoryViewer_ gui plugin, you should see an entry for your memory (`My`) and the core segment (`Data` under `My`).

# Conclusion

That's it! You now have a memory server providing a place where your nice and interesting data can feel welcome, and where interested listeners would look to find it.

[Next, we learn how to commit data to the memory, and query data from the memory.](ArMem/How-to-read-from-and-write-to-a-Memory-Server)