/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::ObjectInstanceToIndex
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/time/Frequency.h>

#include <ArmarXGui/libraries/ArmarXGuiComponentPlugins/LightweightRemoteGuiComponentPlugin.h>

#include <RobotAPI/libraries/ArmarXObjects/plugins/ObjectPoseClientPlugin.h>
#include <RobotAPI/libraries/armem/client/Reader.h>
#include <RobotAPI/libraries/armem/client/Writer.h>
#include <RobotAPI/libraries/armem/client/plugins/ListeningPluginUser.h>

#include "impl/ObjectInstanceToIndex.h"


namespace armarx
{

    /**
     * @defgroup Component-ObjectInstanceToIndex ObjectInstanceToIndex
     * @ingroup RobotAPI-Components
     *
     * Links object instances in the "Index/Spatial" and "Index/Named" memory segments.
     *
     * @class ObjectInstanceToIndex
     * @ingroup Component-ObjectInstanceToIndex
     * @brief Brief description of class ObjectInstanceToIndex.
     *
     * Links object instances in the "Index/Spatial" and "Index/Named" memory segments.
     */
    // Class is currently called like component for compatibility with legacy CMake.
    class ObjectInstanceToIndex :
        virtual public armarx::Component,
        virtual public armarx::armem::ListeningClientPluginUser,
        virtual public armarx::LightweightRemoteGuiComponentPluginUser
    {
    public:
        ObjectInstanceToIndex();

        /// @see armarx::ManagedIceObject::getDefaultName()
        std::string getDefaultName() const override;

    protected:
        armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;

        void onInitComponent() override;
        void onConnectComponent() override;
        void onDisconnectComponent() override;
        void onExitComponent() override;


        // LightweightRemoteGuiComponentPluginUser interface
    public:
        void createRemoteGuiTab();
        void RemoteGui_update() override;


    private:
        void processRobotState(const armem::MemoryID& id,
                               const std::vector<armem::MemoryID>& snapshotIDs);
        void processObjectInstance(const armem::MemoryID& id,
                                   const std::vector<armem::MemoryID>& snapshotIDs);


    private:
        struct Properties
        {
            struct Object
            {
                float maxFrequencyHz = 10;
            };
            Object object;
        };
        Properties properties;

        armem::client::Writer indexSpatialMemoryWriter;


        armarx::plugins::ObjectPoseClientPlugin* objectClientPlugin = nullptr;

        std::mutex objectMutex;
        std::optional<armem::objects::ObjectInstanceToIndex> object;

        armem::client::Reader robotMemoryReader;

        // armarx::RunningTask<ObjectInstanceToIndex>::pointer_type task;

        struct RemoteGuiTab : RemoteGui::Client::Tab
        {
        };
        RemoteGuiTab tab;
    };
} // namespace armarx
